module Collision ( collisionTankBorder
                 , collisionBulletBullet
                 , collisionBulletTank
                 ) where

import Tank
import Game

import Data.Fixed
import Data.Ratio

tankWidth :: Micro
tankWidth = 0.4

tankLength :: Micro
tankLength = 0.95

bulletDiameter :: Micro
bulletDiameter = 0.05

collisionTankBorder :: Micro -> Micro -> Tank -> Tank
collisionTankBorder lw lh tank = tank {tankX = newx, tankY = newy}
    where
      dir = (fromRational . toRational . tankDir $ tank)*pi/180
      cosd = fromRational (round ((cos dir)*1000000)%1000000)
      sind = fromRational (round ((sin dir)*1000000)%1000000)
      
      points = [ (tankLength/2, tankWidth/2)
               , (-tankLength/2, tankWidth/2)
               , (-tankLength/2, -tankWidth/2)
               , (tankLength/2, -tankWidth/2)
               ]
      
      rotp (x, y) = (cosd*x - sind*y, sind*x + cosd*y)
      transp (x, y) = (x + tankX tank, y + tankY tank)
      
      pointst = map (transp . rotp) points
      minx = minimum $ map fst pointst
      maxx = maximum $ map fst pointst
      miny = minimum $ map snd pointst
      maxy = maximum $ map snd pointst
      
      dx = if minx < 0 then (-minx) else if maxx > lw then (-maxx+lw) else 0
      dy = if miny < 0 then (-miny) else if maxy > lh then (-maxy+lh) else 0
      
      newx = (tankX tank) + dx
      newy = (tankY tank) + dy

collisionBulletBullet :: (Bullet, Bullet) -> (Bullet, Bullet) -> Bool
collisionBulletBullet (b1, b1') (b2, b2') = distancesq < (bulletDiameter^2)
    where
      distancesq = (bulletX b1' - bulletX b2')^2 + (bulletY b1' - bulletY b2')^2

collisionBulletTank :: (Bullet, Bullet) -> (Tank, Tank) -> Bool
collisionBulletTank (b, b') (tank, tank') = (not ((between bx minx maxx) && (between by miny maxy))) && ((between bx' minx maxx) && (between by' miny maxy))
    where
      between x a b = x >= a && x <= b
      
      dir t = (fromRational . toRational . tankDir $ t)*pi/180
      cosd t = fromRational (round ((cos $ dir t)*1000000)%1000000)
      sind t = fromRational (round ((sin $ dir t)*1000000)%1000000)
      
      rotp t (x, y) = ((cosd t)*x + (sind t)*y, -(sind t)*x + (cosd t)*y)
      transp t (x, y) = (x - tankX t, y - tankY t)
      
      (bx, by) = (rotp tank) . (transp tank) $ (bulletX b, bulletY b)
      (bx', by') = (rotp tank') . (transp tank') $ (bulletX b', bulletY b')
      
      minx = -tankLength/2
      maxx = tankLength/2
      miny = -tankWidth/2
      maxy = tankWidth/2
